import asyncio
from collections.abc import AsyncGenerator

import pytest_asyncio
import uvicorn
from httpx import AsyncClient

from progtc.server.api import app


@pytest_asyncio.fixture
async def test_server() -> AsyncGenerator[str, None]:
    config = uvicorn.Config(app, host="127.0.0.1", port=0, log_level="warning")
    server = uvicorn.Server(config)
    task = asyncio.create_task(server.serve())
    while not server.started:
        await asyncio.sleep(0.01)
    port = server.servers[0].sockets[0].getsockname()[1]
    yield f"http://127.0.0.1:{port}"
    server.should_exit = True
    await task


@pytest_asyncio.fixture
async def test_client(test_server: str) -> AsyncGenerator[AsyncClient, None]:
    async with AsyncClient(
        base_url=test_server,
        headers={"X-Progtc-API-Key": "1234567890"},
    ) as client:
        yield client



import textwrap

import pytest
from httpx import AsyncClient
from httpx_sse import aconnect_sse

from progtc.types import MessageType


@pytest.mark.asyncio
async def test_execute_code_basic(test_client: AsyncClient) -> None:
    code = textwrap.dedent(
        """\
        async def main():
            print("main function called")
        await main()
        """,
    )
    messages = []
    async with aconnect_sse(
        test_client,
        "POST",
        "/execute-code",
        json={"tool_names": [], "code": code},
    ) as event_source:
        async for event in event_source.aiter_sse():
            messages.append(event.json())
    assert messages[0]["type"] == MessageType.SUCCESS
    assert messages[0]["stdout"] == "main function called\n"


@pytest.mark.asyncio
async def test_execute_code_with_func_call(test_client: AsyncClient) -> None:
    code = textwrap.dedent(
        """\
        async def say_hello():
            print("say_hello function called")
        async def main():
            print("main function called")
            await say_hello()
        await main()
        """,
    )
    messages = []
    async with aconnect_sse(
        test_client,
        "POST",
        "/execute-code",
        json={"tool_names": [], "code": code},
    ) as event_source:
        async for event in event_source.aiter_sse():
            messages.append(event.json())
    assert messages[0]["type"] == MessageType.SUCCESS
    assert messages[0]["stdout"] == "main function called\nsay_hello function called\n"


@pytest.mark.asyncio
async def test_execute_code_with_tool_call(test_client: AsyncClient) -> None:
    code = textwrap.dedent(
        """\
        from tools import get_weather
        async def main():
            print("main function called")
            return await get_weather()
        print(await main())
        """,
    )
    messages = []
    async with aconnect_sse(
        test_client,
        "POST",
        "/execute-code",
        json={"tool_names": ["get_weather"], "code": code},
        timeout=10,
    ) as event_source:
        async for event in event_source.aiter_sse():
            message = event.json()
            messages.append(message)
            if (
                message["type"] == MessageType.TOOL_CALL
                and message["tool_name"] == "get_weather"
            ):
                await test_client.post(
                    "/tool-results",
                    json={
                        "execution_id": message["execution_id"],
                        "tool_call_id": message["id"],
                        "result": "Sunny",
                    },
                )
    assert messages[0]["type"] == MessageType.TOOL_CALL, messages[0]
    assert messages[0]["tool_name"] == "get_weather"
    assert messages[0]["args"] == []
    assert messages[0]["kwargs"] == {}
    assert messages[1]["type"] == MessageType.SUCCESS
    assert messages[1]["stdout"] == "main function called\nSunny\n"


@pytest.mark.asyncio
async def test_execute_code_with_tool_call_with_args(test_client: AsyncClient) -> None:
    code = textwrap.dedent(
        """\
        from tools import get_weather
        async def main():
            from tools import get_weather
            print("main function called")
            return await get_weather("UK", city="London")
        print(await main())
        """,
    )
    messages = []
    async with aconnect_sse(
        test_client,
        "POST",
        "/execute-code",
        json={"tool_names": ["get_weather"], "code": code},
        timeout=10,
    ) as event_source:
        async for event in event_source.aiter_sse():
            message = event.json()
            messages.append(message)
            if (
                message["type"] == MessageType.TOOL_CALL
                and message["tool_name"] == "get_weather"
            ):
                await test_client.post(
                    "/tool-results",
                    json={
                        "execution_id": message["execution_id"],
                        "tool_call_id": message["id"],
                        "result": "Sunny",
                    },
                )
    assert messages[0]["type"] == MessageType.TOOL_CALL, messages[0]
    assert messages[0]["tool_name"] == "get_weather"
    assert messages[0]["args"] == ["UK"]
    assert messages[0]["kwargs"] == {"city": "London"}
    assert messages[1]["type"] == MessageType.SUCCESS
    assert messages[1]["stdout"] == "main function called\nSunny\n"



from __future__ import annotations

import ast
import asyncio
import multiprocessing as mp
import queue
import sys
import traceback
from collections.abc import Generator
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from functools import partial
from io import StringIO
from multiprocessing import Queue as MPQueue
from multiprocessing.queues import Queue as MPQueueType
from types import ModuleType
from uuid import uuid4

from progtc.server.tool_call_manager import ToolCallManager
from progtc.types import (
    CodeRuntimeError,
    CodeSyntaxError,
    ExecuteCodeError,
    ExecuteCodeSuccess,
    ToolCall,
    ToolCallResult,
)


@contextmanager
def _capture_stdout_stderr() -> Generator[tuple[StringIO, StringIO], None, None]:
    stdout = StringIO()
    stderr = StringIO()
    with redirect_stdout(stdout), redirect_stderr(stderr):
        yield stdout, stderr


async def execute_code(
    execution_id: str,
    code: str,
    tool_names: list[str],
    tool_call_queue: MPQueueType[ToolCall],
    tool_call_results_queue: MPQueueType[ToolCallResult],
    code_execution_result_queue: MPQueueType[ExecuteCodeSuccess | ExecuteCodeError],
) -> None:
    async with ToolCallManager(
        execution_id=execution_id,
        tool_call_queue=tool_call_queue,
        tool_call_results_queue=tool_call_results_queue,
    ) as tool_call_manager:
        # Create a tools module which the code can import tools from
        # and make it availabl in the sys.modules dictionary.
        tools_module = ModuleType("tools")
        for tool_name in tool_names:
            setattr(
                tools_module,
                tool_name,
                partial(
                    tool_call_manager.call_tool,
                    tool_name,
                ),
            )
        sys.modules["tools"] = tools_module

        # Run the code while capturing stdout and stderr.
        with _capture_stdout_stderr() as (stdout, stderr):
            try:
                compiled = compile(
                    code,
                    "<string>",
                    "exec",
                    flags=ast.PyCF_ALLOW_TOP_LEVEL_AWAIT,
                )
            except (SyntaxError, ValueError):
                traceback.print_exc()
                code_execution_result_queue.put(
                    CodeSyntaxError(
                        stdout=stdout.getvalue(),
                        stderr=stderr.getvalue(),
                    ),
                )
                return

            try:
                exec_globals = {"__builtins__": __builtins__}
                coro = eval(compiled, exec_globals)
                if asyncio.iscoroutine(coro):
                    await coro
            except Exception:
                traceback.print_exc()
                code_execution_result_queue.put(
                    CodeRuntimeError(
                        stdout=stdout.getvalue(),
                        stderr=stderr.getvalue(),
                    ),
                )
                return

    code_execution_result_queue.put(
        ExecuteCodeSuccess(stdout=stdout.getvalue(), stderr=stderr.getvalue()),
    )


def execute_code_sync(
    execution_id: str,
    code: str,
    tool_names: list[str],
    tool_call_queue: MPQueueType[ToolCall],
    tool_call_results_queue: MPQueueType[ToolCallResult],
    code_execution_result_queue: MPQueueType[ExecuteCodeSuccess | ExecuteCodeError],
) -> None:
    asyncio.run(
        execute_code(
            execution_id=execution_id,
            code=code,
            tool_names=tool_names,
            tool_call_queue=tool_call_queue,
            tool_call_results_queue=tool_call_results_queue,
            code_execution_result_queue=code_execution_result_queue,
        ),
    )


class CodeExecutionProcess:
    def __init__(self, code: str, tool_names: list[str]) -> None:
        self.execution_id = uuid4().hex
        self._tool_call_queue: MPQueueType[ToolCall] = MPQueue()
        self._tool_call_results_queue: MPQueueType[ToolCallResult] = MPQueue()
        self._code_execution_result_queue: MPQueueType[
            ExecuteCodeSuccess | ExecuteCodeError
        ] = MPQueue()
        self._process = mp.Process(
            target=execute_code_sync,
            kwargs=dict(
                execution_id=self.execution_id,
                code=code,
                tool_names=tool_names,
                tool_call_queue=self._tool_call_queue,
                tool_call_results_queue=self._tool_call_results_queue,
                code_execution_result_queue=self._code_execution_result_queue,
            ),
        )

    def start(self) -> None:
        self._process.start()

    def stop(self) -> None:
        if self._process.is_alive():
            self._process.terminate()
        self._process.join()

    def is_done(self) -> bool:
        return self._process.exitcode is not None

    def result(self) -> ExecuteCodeSuccess | ExecuteCodeError:
        if not self.is_done():
            raise RuntimeError("Code execution process not done")
        return self._code_execution_result_queue.get(timeout=0.1)

    def next_tool_call(self, timeout: float = 0.1) -> ToolCall | None:
        try:
            return self._tool_call_queue.get(timeout=timeout)
        except queue.Empty:
            return None

    def send_tool_result(self, tool_call_result: ToolCallResult) -> None:
        self._tool_call_results_queue.put(tool_call_result)



import secrets
import time
from collections.abc import Generator

from fastapi import APIRouter, Depends, FastAPI, Header, HTTPException, Response
from fastapi.responses import StreamingResponse

from progtc.server.code_execution_manager import CodeExecutionManager
from progtc.server.config import server_config
from progtc.types import (
    CodeTimeoutError,
    ExecuteCodeRequest,
    ToolCallResult,
)

code_execution_manager = CodeExecutionManager()


def authenticate(
    x_progtc_api_key: str = Header(),
) -> None:
    if not secrets.compare_digest(x_progtc_api_key, server_config.api_key):
        raise HTTPException(status_code=401, detail="Unauthorized")


app = FastAPI()


@app.get("/ping")
def ping() -> Response:
    return Response(status_code=200)


router = APIRouter(dependencies=[Depends(authenticate)])


@router.post("/tool-results")
def add_tool_result(tool_call_result: ToolCallResult) -> None:
    code_execution_manager.send_tool_result(tool_call_result)


@router.post("/execute-code")
def execute_code(body: ExecuteCodeRequest) -> StreamingResponse:
    def stream() -> Generator[str, None, None]:
        with code_execution_manager.run(body.code, body.tool_names) as process:
            t0 = time.monotonic()

            # Yield tool calls to the client until the process
            # is done or the timeout is reached
            while (
                not process.is_done()
                and time.monotonic() - t0 < server_config.code_execution_timeout
            ):
                tool_call = process.next_tool_call()
                if tool_call is None:
                    continue
                yield f"data: {tool_call.model_dump_json()}\n\n"

            # Yield the final result to the client.
            if process.is_done():
                result = process.result()
            else:
                result = CodeTimeoutError(
                    stdout="",
                    stderr="",
                )
            yield f"data: {result.model_dump_json()}\n\n"

    return StreamingResponse(stream(), media_type="text/event-stream")


app.include_router(router)



from __future__ import annotations

import asyncio
import queue
from multiprocessing.queues import Queue as MPQueueType
from types import TracebackType

from pydantic import JsonValue

from progtc.server.config import server_config
from progtc.types import (
    ToolCall,
    ToolCallResult,
)


class ToolCallManager:
    def __init__(
        self,
        execution_id: str,
        tool_call_queue: MPQueueType[ToolCall],
        tool_call_results_queue: MPQueueType[ToolCallResult],
    ) -> None:
        self._execution_id = execution_id
        self._tool_call_queue = tool_call_queue
        self._tool_call_results_queue = tool_call_results_queue
        self._tool_call_results_split_queue: dict[
            str,
            asyncio.Queue[ToolCallResult],
        ] = {}
        self._handle_results_task: asyncio.Task[None] | None = None
        self._shutdown_event = asyncio.Event()

    async def _handle_results_queue(self) -> None:
        while not self._shutdown_event.is_set():
            try:
                tool_call_result = self._tool_call_results_queue.get_nowait()
                await self._tool_call_results_split_queue[
                    tool_call_result.tool_call_id
                ].put(tool_call_result)
            except queue.Empty:
                await asyncio.sleep(0.1)

    async def __aenter__(self) -> ToolCallManager:
        self._handle_results_task = asyncio.create_task(self._handle_results_queue())
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self._shutdown_event.set()
        if self._handle_results_task is not None:
            await self._handle_results_task

    async def call_tool(
        self,
        tool_name: str,
        *args: JsonValue,
        **kwargs: JsonValue,
    ) -> JsonValue:
        tool_call = ToolCall(
            execution_id=self._execution_id,
            tool_name=tool_name,
            args=args,
            kwargs=kwargs,
        )
        self._tool_call_results_split_queue[tool_call.id] = asyncio.Queue[
            ToolCallResult
        ]()
        self._tool_call_queue.put(tool_call)
        try:
            result = await asyncio.wait_for(
                self._tool_call_results_split_queue[tool_call.id].get(),
                timeout=server_config.tool_call_timeout,
            )
            return result.result
        finally:
            del self._tool_call_results_split_queue[tool_call.id]



import asyncio
import json
import secrets

from dotenv import load_dotenv
from e2b_code_interpreter import AsyncSandbox
from pydantic_ai import Agent
from pydantic_ai.messages import (
    ModelRequest,
    ModelResponse,
    SystemPromptPart,
    TextPart,
    ToolCallPart,
    ToolReturnPart,
    UserPromptPart,
)
from rich.console import Console
from rich.panel import Panel
from rich.syntax import Syntax

from progtc import AsyncProgtcClient

load_dotenv()

PROGTC_API_KEY = secrets.token_urlsafe(32)
PROGTC_PORT = 8000


async def get_weather(city: str, country: str) -> str:
    """Get the weather for a city."""
    return f"Weather in {city}, {country}: Rainy"


async def get_reviews(city: str, country: str) -> str:
    """Get reviews for a city."""
    return f"Reviews in {city}, {country}: 4.5 stars"


async def execute_code_with_tools(code: str) -> dict[str, str]:
    """Execute python code in a sandboxed environment and return the stdout and stderr.

    You have the following available tools:
    - async def get_weather(city: str, country: str) -> str:
    - async def get_reviews(city: str, country: str) -> str:

    Guidelines:
    - All tools are async and must be awaited.
    - You must import the tools from the tools module.
    - Try to consolidate your requests into a single script for efficiency.
    - The code runs in a top-level async context.
    - You can use `await` directly without defining an async function.
    - Print the data you want to be returned in stdout.

    Example Code:
    ```python
    from tools import get_weather
    weather = await get_weather("London", "UK")
    print(f"The weather in London, UK is {weather}")
    ````

    """
    try:
        # Start e2b sandbox
        sandbox = await AsyncSandbox.create()

        # Install progtc server in the sandbox
        await sandbox.commands.run("pip install 'progtc[server]'")

        # Start the progtc server in the sandbox and get url
        await sandbox.commands.run(
            f"progtc serve --host 0.0.0.0 --port {PROGTC_PORT} "
            f"--api-key {PROGTC_API_KEY}",
            background=True,
        )
        host = sandbox.get_host(PROGTC_PORT)

        # Create client and execute code
        client = AsyncProgtcClient(
            base_url=f"https://{host}",
            api_key=PROGTC_API_KEY,
        )
        result = await client.execute_code(
            code=code,
            tools={
                "get_weather": get_weather,
                "get_reviews": get_reviews,
            },
        )

        return {"stdout": result.stdout, "stderr": result.stderr}

    finally:
        await sandbox.kill()


async def main() -> None:
    agent = Agent(
        "openai:gpt-5.1",
        system_prompt=(
            "You are a helpful assistant that can execute python code "
            "to answer questions."
        ),
        tools=[execute_code_with_tools],
    )

    result = await agent.run(
        "Can you get the weather and reviews score for "
        "London, Tokyo and Paris using your code execution tools?"
    )

    # Print the conversation
    console = Console()
    for msg in result.new_messages():
        if isinstance(msg, ModelRequest):
            for request_part in msg.parts:
                if isinstance(request_part, SystemPromptPart):
                    console.print(
                        Panel(request_part.content, title="System", border_style="dim")
                    )
                elif isinstance(request_part, UserPromptPart):
                    console.print(
                        Panel(
                            str(request_part.content), title="User", border_style="blue"
                        )
                    )
                elif isinstance(request_part, ToolReturnPart):
                    content = (
                        request_part.content
                        if isinstance(request_part.content, dict)
                        else json.loads(request_part.content)
                    )
                    console.print(
                        Panel(
                            f"stdout:\n\n{content['stdout']}\n\nstderr:\n\n{content['stderr']}",
                            title=f"Tool Return: {request_part.tool_name}",
                            border_style="green",
                        )
                    )
        elif isinstance(msg, ModelResponse):
            for response_part in msg.parts:
                if isinstance(response_part, TextPart):
                    console.print(
                        Panel(
                            response_part.content,
                            title="Assistant",
                            border_style="magenta",
                        )
                    )
                elif isinstance(response_part, ToolCallPart):
                    args = (
                        response_part.args
                        if isinstance(response_part.args, dict)
                        else json.loads(response_part.args or '{"code": ""}')
                    )
                    code = Syntax(
                        args["code"], "python", theme="monokai", line_numbers=True
                    )
                    console.print(
                        Panel(
                            code,
                            title=f"Tool Call: {response_part.tool_name}",
                            border_style="yellow",
                        )
                    )


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import time

from httpx import AsyncClient
from httpx_sse import aconnect_sse
from pydantic import TypeAdapter

from progtc.types import (
    ExecuteCodeError,
    ExecuteCodeMessage,
    ExecuteCodeSuccess,
    ToolCall,
    ToolHandler,
)

ExecuteCodeMessageAdapter = TypeAdapter[ExecuteCodeMessage](ExecuteCodeMessage)


DEFAULT_PING_TIMEOUT = 5.0
DEFAULT_WAIT_FOR_SERVER_TIMEOUT = 10.0
DEFAULT_CODE_EXECUTION_TIMEOUT = 30.0
DEFAULT_TOOL_CALL_RESULT_TIMEOUT = 5.0


class AsyncProgtcClient:
    def __init__(
        self,
        base_url: str,
        api_key: str,
        headers: dict[str, str] | None = None,
    ):
        self._base_url = base_url
        self._api_key = api_key
        self._headers = headers or {}

    async def ping(
        self,
        timeout: float | None = DEFAULT_PING_TIMEOUT,
        headers: dict[str, str] | None = None,
    ) -> bool:
        merged_headers = {**self._headers, **(headers or {})}
        async with AsyncClient(
            base_url=self._base_url, headers=merged_headers, timeout=timeout
        ) as client:
            try:
                response = await client.get("/ping")
                return response.status_code == 200
            except Exception:
                return False

    async def wait_for_server(
        self,
        timeout: float | None = DEFAULT_WAIT_FOR_SERVER_TIMEOUT,
        headers: dict[str, str] | None = None,
    ) -> None:
        t0 = time.monotonic()
        ping_timeout = (
            min(timeout, DEFAULT_PING_TIMEOUT)
            if timeout is not None
            else DEFAULT_PING_TIMEOUT
        )
        while not await self.ping(timeout=ping_timeout, headers=headers):
            await asyncio.sleep(0.5)
            if timeout is not None and time.monotonic() - t0 > timeout:
                raise TimeoutError(f"Progtc server not ready after {timeout}s")

    async def execute_code(
        self,
        code: str,
        tools: dict[str, ToolHandler],
        timeout: float | None = DEFAULT_CODE_EXECUTION_TIMEOUT,
        headers: dict[str, str] | None = None,
    ) -> ExecuteCodeSuccess | ExecuteCodeError:
        merged_headers = {
            **self._headers,
            **(headers or {}),
            "X-Progtc-API-Key": self._api_key,
        }
        async with (
            AsyncClient(
                base_url=self._base_url,
                headers=merged_headers,
                timeout=timeout,
            ) as client,
            aconnect_sse(
                client,
                "POST",
                "/execute-code",
                json={
                    "tool_names": list(tools.keys()),
                    "code": code,
                },
            ) as event_source,
        ):

            async def handle_tool_call(tool_call: ToolCall) -> None:
                handler = tools[tool_call.tool_name]
                result = await handler(*tool_call.args, **tool_call.kwargs)
                await client.post(
                    "/tool-results",
                    json={
                        "execution_id": tool_call.execution_id,
                        "tool_call_id": tool_call.id,
                        "result": result,
                    },
                    timeout=DEFAULT_TOOL_CALL_RESULT_TIMEOUT,
                )

            pending_tasks: list[asyncio.Task[None]] = []
            async for event in event_source.aiter_sse():
                event_obj = ExecuteCodeMessageAdapter.validate_python(event.json())
                if isinstance(event_obj, ToolCall):
                    task = asyncio.create_task(handle_tool_call(event_obj))
                    pending_tasks.append(task)
                else:
                    if pending_tasks:
                        await asyncio.gather(*pending_tasks)
                    return event_obj

        raise RuntimeError("Stream ended without result")
